#!/usr/bin/perl -w

###########################################################################
# 
#  wneditor.pl - a WordNet editor
#  allows editing, adding, deleting and linking of synsets
#
#  developed as a part of the "AudioClas" project
#  Copyright (C) 2002 MTG / IUA / UPF, Barcelona, Spain
#  Written by koppi@iua.upf.es
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#  
###########################################################################

######
# TODO 
# - hmm, maybe we should not use %in but local vars.
# - check if there is already a pointer between two synsets
#   a holonym of a synsset must not be a hypernym

# for development only -> remove me!!
BEGIN {
	@INC = ("/work/src/Lingua-Wordnet/blib/lib/", @INC);
}


#####
# CONFIGURATION
####
my %conf = (
	"wndir" => "/tmp/lwn",			# full path to your Lingua::Wordnet database
	"webdir" => "/wneditor/",		# base-url of your css and images
	"iconwidth" => "16",			# width and height of your icons
	"iconheight" => "16",
	"ispell" => "/usr/bin/ispell",  # path to your ispell binary
);
# END of configuration 

use strict;
use diagnostics;

use CGI;

use constant DEBUG => 0;

use HTML::Entities;

use Lingua::Wordnet;

use Lingua::Stem;
use Lingua::Ispell qw(spellcheck);
Lingua::Ispell::allow_compounds(1);
$Lingua::Ispell::path = '/usr/bin/ispell';

my $stemmer;
our @types = ('n', 'v', 'a', 's', 'r');
my %types = (
	'n'	=> 'Noun(s)', 
	'v'	=> 'Verb(s)', 
	'a'	=> 'Adjective(s)', 
	's'	=> 'Adjective Satellite(s)', 
	'r' => 'Adverbs(s)'
);

my @datatypes = ("antonyms","hypernyms","hyponyms",
    "comp_meronyms","member_meronyms","stuff_meronyms","portion_meronyms",
	"feature_meronyms","place_meronyms","phase_meronyms",
    "comp_holonyms","member_holonyms","stuff_holonyms","portion_holonyms",
    "feature_holonyms","place_holonyms","phase_holonyms",
	"category_domains", "region_domains", "usage_domains",
	"category_domainterms", "region_domainterms", "usage_domainterms",
	"derivedforms",
    "attributes","entailment","causes","pertainyms");

my %datatypes = (
	"antonyms"			=> "Antonyms",
	"hypernyms"			=> "Hypernyms",
	"hyponyms"			=> "Hyponyms",
	"entailment"		=> "Entailment",
	"comp_meronyms"		=> "Component Meronyms",
	"member_meronyms"	=> "Member Meronyms",
	"stuff_meronyms"	=> "Stuff Meronyms",
	"portion_meronyms"	=> "Portion Meronyms",
	"feature_meronyms"	=> "Feature Meronyms",
	"place_meronyms"	=> "Place Meronyms",
	"phase_meronyms"	=> "Phase Meronyms",
	"comp_holonyms"		=> "Comp Holonyms",
	"member_holonyms"	=> "Member Holonyms",
	"stuff_holonyms"	=> "Stuff Holonyms",
	"portion_holonyms"	=> "Portion Holonyms",
	"feature_holonyms"	=> "Feature Holonyms",
	"place_holonyms"	=> "Place Holonyms",
	"phase_holonyms"	=> "Phase Holonyms",
	"attributes"		=> "Attributes",
	"category_domains"	=> "Category Domains",
	"region_domains"	=> "Region Domains",
	"usage_domains"		=> "Usage Domains",
	"category_domainterms"=> "Category Domain Terms",
	"region_domainterms"=> "Region Domain Terms",
	"usage_domainterms"	=> "Usage Domain Terms",
	"derivedforms"		=> "Derived Forms",
	#"functions"			=> "Functions",
	"causes"			=> "Causes",
	"pertainyms"		=> "Pertainyms"
);

my %datatypedefinitions = (
	"antonyms"			=> "A word of opposite meaning (the usual antonym of 'good' is 'bad').",
	"hypernyms"			=> "The generic term used to designate a whole class of specific instances. Y is a hypernym of X if X is a (kind of) Y.",
	"hyponyms"			=> "The specific term used to designate a member of a class. X is a hyponym of Y if X is a (kind of) Y.",
	"entailment"		=> "A verb X entails Y if X cannot be done unless Y is, or has been, done.",
	"comp_meronyms"		=> "The name of a constituent part of something. X is a part of Y.",
	"member_meronyms"	=> "A member of something. X is a member of group Y.",
	"stuff_meronyms"	=> "The substance of something. Y is made of X",
	"portion_meronyms"	=> "A segment of something. X is a segment or portion of Y",
	"feature_meronyms"	=> "A feature of something. X is part of the activity Y",
	"place_meronyms"	=> "A place somewhere. X is a place in Y",
	"phase_meronyms"	=> "A part of process. X is part of the process Y",
	"comp_holonyms"		=> "The name of the whole of which the meronym names a part. Y is a holonym of X if X is a part of Y.",
	"member_holonyms"	=> "Y's are members of group X",
	"stuff_holonyms"	=> "X is made of Y'",
	"portion_holonyms"	=> "",
	"feature_holonyms"	=> "",
	"place_holonyms"	=> "",
	"phase_holonyms"	=> "",
	"attributes"		=> "A noun for which adjectives express values. The noun weight  is an attribute, for which the adjectives light  and heavy  express values.",
	"category_domains"	=> "A topical classification (category) to which a synset has been linked",
	"region_domains"	=> "A topical classification (region) to which a synset has been linked",
	"usage_domains"		=> "A topical classification (usage) to which a synset has been linked",
	"category_domainterms"=> "A synset belonging to a topical class (category).",
	"region_domainterms"=> "A synset belonging to a topical class (region).",
	"usage_domainterms"	=> "A synset belonging to a topical class (usage).",
#	"functions"			=> "",
	"derivedforms"		=> "Links for derivational morphology",
	"causes"			=> "",
	"pertainyms"		=> "A relational adjective. Adjectives that are pertainyms are usually defined by such phrases as 'of or pertaining to' and do not have antonyms. A pertainym can point to a noun or another pertainym."
);

my %reversepointer = (
	"antonyms"			=> "antonyms",
	"hypernyms"			=> "hyponyms",
	"hyponyms"			=> "hypernyms",
	"comp_meronyms"		=> "comp_holonyms",
	"member_meronyms"	=> "member_holonyms",
	"stuff_meronyms"	=> "stuff_holonyms",
	"portion_meronyms"	=> "portion_holonyms",
	"feature_meronyms"	=> "feature_holonyms",
	"place_meronyms"	=> "place_holonyms",
	"phase_meronyms"	=> "phase_holonyms",
	"comp_holonyms"		=> "comp_meronyms",
	"member_holonyms"	=> "member_meronyms",
	"stuff_holonyms"	=> "stuff_meronyms",
	"portion_holonyms"	=> "portion_meronyms",
	"feature_holonyms"	=> "feature_meronyms",
	"place_holonyms"	=> "place_meronyms",
	"phase_holonyms"	=> "phase_meronyms",
	"category_domains"	=> "category_domainterms",
	"region_domains"	=> "region_domainterms",
	"usage_domains"		=> "usage_domainterms",
	"category_domainterms"=> "category_domains",
	"region_domainterms"=> "region_domains",
	"usage_domainterms"	=> "usage_domains",
	"derivedforms"		=> "derivedforms",

#	"attributes"		=> "Attributes",
#	"causes"			=> "Causes",
#	"pertainyms"		=> "Pertainyms"
);


my $cgi = new CGI;

print $cgi->header();

# get http params
my %in = getParams(QUERY => $cgi);

# who am i?
my $url = $cgi->url();

# dump input params to STDERR for debug
dprint(\%in) if DEBUG;

my $html .= &head;

# we will show the searchform at every page
$html .= &searchform;

my $wn = new Lingua::Wordnet("$conf{wndir}");

# show a synset (by offset)
if ($in{show}) {
	$html .= &show_synset($in{show}, $in{open});

# save words
} elsif ($in{save}) {
	$html .= &save($in{edit}, $in{words}, $in{pos}, $in{gloss});

# edit words and glossary (form)
} elsif ($in{edit}) {

	# do something (del/add/... words)
	if ($in{ACTION_NAME}) {
		# delete word from synset
		if ($in{ACTION_NAME} eq "delword") {
			my $num = $in{ACTION_ARGV}->[0];
			splice (@{$in{words}}, $num, 1);
		}
		# move word up
		if ($in{ACTION_NAME} eq "upword") {
			my $num = $in{ACTION_ARGV}->[0];
			@{$in{words}}[$num-1, $num] = @{$in{words}}[$num, $num-1];
		}
		# move word down
		if ($in{ACTION_NAME} eq "downword") {
			my $num = $in{ACTION_ARGV}->[0];
			@{$in{words}}[$num, $num+1] = @{$in{words}}[$num+1, $num];
		}
		# add a word 
		if ($in{ACTION_NAME} eq "addword") {
			push @{$in{words}}, "";
		}
	}
	$html .= &edit_form($in{edit}, $in{words}, $in{pos}, $in{gloss});


# delete synset (ask if it really should be deleted) 
} elsif ($in{del}) {
	$html .= &del_form($in{del});

# really delete synset and all pointers 
} elsif ($in{dodel}) {
	$html .= &delete($in{dodel});

# edit pointer to datatype (form)
} elsif ($in{editpointer}) {
	$html .= &pointer_form($in{editpointer}, $in{offset}, $in{q});

# save pointer
} elsif ($in{savepointer}) {
	$html .= &save_pointer($in{savepointer}, $in{offset}, $in{addoffset});

# delete pointer
} elsif ($in{delpointer}) {
	$html .= &del_pointer($in{delpointer}, $in{offset}, $in{deloffset});

# add a new synset (form) 
} elsif ($in{addsynset}) {
	$html .= &add_form($in{words}, $in{pos}, $in{gloss}, 0, "");

# lookup a new synset
} elsif ($in{lookupnewsynset}) {
	$html .= &add_form($in{words}, $in{pos}, $in{gloss}, 1, "");

# save a new synset
} elsif ($in{savesynset}) {
	$html .= &save_synset($in{words}, $in{pos}, $in{gloss}, $in{really}, $in{oldwords}, $in{oldpos});

# search for synsets
} elsif ($in{q}) {
	$html .= &search($in{q});
}

$html .= &foot;
print $html;

$wn->close;

sub show_synset {
	my $offset = shift;
	my $open = shift;

#	my $wn = new Lingua::Wordnet("$conf{wndir}");

	my $synset = get_synset_by_offset($wn, $offset);

	if (!$synset) {
		return qq|<div class="error">No Synset found with offset '$offset'<br/>
	[<a href="javascript:history.back();">back</a>]
</div>|;
	}

	my ($txt, $gloss) = get_text_gloss($synset);

	my $html;
	$html .= qq|<div class="box">\n|;
	$html .= qq|\t<div class="boxheader">\n|;
	$html .= qq|\t\t<div style="float:right;">

	<a href="$url?del=$offset"><img src="$conf{webdir}/images/del.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Delete Synset" title="Delete Synset" hspace="10"></a>

	<a href="$url?edit=$offset"><img src="$conf{webdir}/images/edit.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Edit Words" title="Edit Words"></a>
</div>|;
	$html .= qq|Detail View of: <span class="word">$txt</span></div>\n|;
	$html .= qq|\t<div class="boxtext">\n|;
	$html .= qq|\t\t<div><span class="label">Offset:</span> $offset</div>\n|;
	$html .= qq|\t</div>\n|;

	$html .= qq|\t<div class="boxsubheader">\n|;
#	$html .= qq|<div class="editbutton"><a href="$url?editgloss=$offset"><img src="$conf{webdir}/images/edit.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Edit" title="Edit"></a></div>|;
	$html .= qq|<img src="$conf{webdir}/images/trans.png" align="absmiddle" width="$conf{iconwidth}" height="$conf{iconheight}" alt="" border="0">\n|;
	$html .= qq|Glossary</div>\n|;
	$html .= qq|\t<div class="listrow0">\t|;
	$html .= qq|\t<span class="glossary">$gloss</span>|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t<br/>\n|;
	foreach my $datatype (@datatypes) {
		my @synsets;
        eval "\@synsets = \$synset->$datatype";
        if ($@) { die "eval error: $@"; }
		my $i = 0;
		my $syncount = scalar(@synsets);
		$html .= qq|<a name="$datatype"></a>\n|;
		$html .= qq|<div class="boxsubheader">\n|;
		$html .= qq|\t<div style="float:right; margin-left: 10px;"><a href="$url?editpointer=$datatype;offset=$offset"><img src="$conf{webdir}/images/edit.png" align="absmiddle" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Edit $datatypes{$datatype}" title="Edit $datatypes{$datatype}" border="0"></a></div>\n|;
		$html .= qq|\t<div style="float:right">$syncount Synsets</div>\n|;
		if ($syncount) {
			if ($open && $open eq $datatype) {
				$html .= qq|<a href="$url?show=$offset;"><img src="$conf{webdir}/images/collapse.png" align="absmiddle" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Close" title="Close" border="0"></a>|;
			} else {
				$html .= qq|<a href="$url?show=$offset;open=$datatype#$datatype"><img src="$conf{webdir}/images/expand.png" align="absmiddle" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Open" title="Open" border="0"></a>|;
			}
		} else {
			$html .= qq|<img src="$conf{webdir}/images/trans.png" align="absmiddle" width="$conf{iconwidth}" height="$conf{iconheight}" alt="" border="0">|;
		}
		$html .= qq|$datatypes{$datatype}|;
		$html .= qq| - <span class="small">$datatypedefinitions{$datatype}</span>| if ($datatypedefinitions{$datatype});
		$html .= qq|</div>\n|;
		if ($open && $open eq $datatype) {
			foreach my $syn (@synsets) {
				my ($words, $gloss) = get_text_gloss($syn);
				my $offset = $syn->offset;
							$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;
						
#							$html .= qq|<div class="editbutton"><a href="$url?edit=$offset"><img src="$conf{webdir}/images/edit.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Edit" title="Edit"></a></div>|;

							$html .= qq|<a href="$url?show=$offset" class="word">$words</a> <span class="offset">($offset)</span>\n|;
							$html .= qq|<div class="glossary">$gloss</div>\n|;

							$html .= qq|</div>|;

			}
		}
		$html .= qq|<br/>|;
	}

	$html .= qq|</div>\n|;
	return $html;
}

sub add_form {
	my $words = shift || "";
	my $pos = shift || "n";
	my $gloss = shift || "";
	my $lookup = shift || "";
	my $info = shift || "";

	my $html;
	$html .= qq|<div class="box">\n|;
	$html .= qq|\t<div class="boxheader">\n|;
	$html .= qq|\t\tAdd a new Synset\n|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t<br/>\n|;


	$html .= qq|\t<div class="boxtext">\n|;

	if ($info) {
		$html .= qq|<div class="info">$info</div><br/>|;
	}

	$html .= &add_form_form($words, $pos, $gloss, 0);


	if ($lookup) {
		if (my $lookuphtml = &lookup_synsets($words, $pos)) {
			$html .= $lookuphtml;
		} else {
			$html .= qq|<div class="info">No Synsets of type $types{$pos} found with '$words'</div>\n|;
		}
	}

	$html .= qq|\t</div>\n|;
	$html .= qq|</div>\n|;
	return $html;
}

sub lookup_synsets {
	my $words = shift;
	my $pos = shift;
	# check if there are already synsets with these words	
	&initspeller();

	my $html = "";

	my @wordsorig = split(/\s+/, $words);
	my ($found, $synsets) = &get_synsets($wn, \@wordsorig);

	my @syns;
	foreach my $w (keys %{$synsets}) {
		foreach my $syn (@{$synsets->{$w}}) {
			if ($syn->[0] eq $pos) {
				push @syns, $syn->[1];
			}
		}
	}

	if (@syns) {
		$html .= qq|\t</div>\n|;
		$html .= qq|\t<div class="boxsubheader">Available Synsets</div>\n|;
		my $i = 0;
		foreach my $syn (@syns) {
			my ($words, $gloss) = get_text_gloss($syn);
			my $xoffset = $syn->offset;
			$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;
			$html .= qq|<a href="$url?show=$xoffset" class="word">$words</a> <span class="offset">($xoffset)</span>\n|;
			$html .= qq|<div class="glossary">$gloss</div>\n|;
			$html .= qq|</div>|;
		}
	} 
	return $html;
}

sub add_form_form {
	my $words = shift || "";
	my $pos = shift || "n";
	my $gloss = shift || "";
	my $really = shift;

	my $html;

	$html .= qq|\t<form action="$url" method="POST">\n|;

	$html .= qq|\t<table border="0" cellpadding="0" cellspacing="0">\n|;
	$html .= qq|\t<tr>\n|;
	$html .= qq|\t\t<td>Part-of-speech: &nbsp;</td>\n|;
	$html .= qq|\t\t<td colspan="2"><select name="pos">\n|;
	foreach my $p (@types) {
		my $sel = ($pos eq $p) ? " selected " : "";
		$html .= qq|\t\t\t<option value="$p"$sel>$types{$p}</option>\n|;
	}
	$html .= qq|\t\t</select></td>\n|;
	$html .= qq|\t</tr>\n\t<tr>\n|;
	$html .= qq|\t\t<td>Word(s) &nbsp;</td>\n|;
	$html .= qq|\t\t<td width="50"><input type="text" name="words" value="| . encode_entities($words) . qq|" size="50"/></td>\n|;
	$html .= qq|\t\t<td>|;
	$html .= qq|<input type="submit" name="lookupnewsynset" value="Lookup" class="formbutton"/>\n|;
	$html .= qq|</td>\n|;
	$html .= qq|\t</tr>\n\t<tr>\n|;
	$html .= qq|\t\t<td>Glossary &nbsp;</td>\n|;
	$html .= qq|\t\t<td colspan="2"><input type="text" name="gloss" value="| . encode_entities($gloss) . qq|" size="70"/></td>\n|;
	$html .= qq|\t</tr>\n\t<tr>\n|;
	$html .= qq|\t\t<td></td>\n|;
	if ($really) {
		$html .= qq|\t\t<input type="hidden" name="really" value="1"/>\n|;
		$html .= qq|\t\t<input type="hidden" name="oldwords" value="| . encode_entities($words) . qq|"/>\n|;
		$html .= qq|\t\t<input type="hidden" name="oldpos" value="$pos"/>\n|;
		$html .= qq|\t\t<td colspan="2"><input type="submit" name="addsynset" value="Back" class="formbutton"/> \n|;
		$html .= qq|\t\t<input type="submit" name="savesynset" value="Really Save" class="formbutton"/></td>\n|;
	} else {
		$html .= qq|\t\t<td colspan="2"><input type="submit" name="savesynset" value="Save" class="formbutton"/></td>\n|;
	}
	$html .= qq|\t</tr>\n|;
	$html .= qq|\t</table>\n|;

	$html .= qq|\t</form>|;

	return $html;
}

sub save_synset {
	my $words = shift;
	my $pos = shift;
	my $gloss = shift;
	my $really = shift;
	my $oldwords = shift || "";
	my $oldpos = shift || ""; 
	
	my $html;

	$html .= qq|<div class="box">\n|;
	$html .= qq|\t<div class="boxheader">\n|;
	$html .= qq|\t\tAdd a new Synset\n|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t<br/>\n|;

	$html .= qq|\t<div class="boxtext">\n|;

	unless ($pos && $words) {
		$html .= "<b>Part-of-speech</b> and <b>words</b> are required!<br/><br/>";
		$html .= &add_form_form($words, $pos, $gloss);
		return $html;
	}

	# force save (and values are unchanged)
	if ($really
		&& $pos eq $oldpos
		&& $words eq $oldwords) {

		# everything fine - we can add the new synset
		$html .= &do_save_synset($pos, $words, $gloss);

	} else {

		if (my $lookuphtml = &lookup_synsets($words, $pos)) {
			$html .= qq|<div class="info">There are already Synsets with this words. Do you really want to add a new synset?</div><br/>|;
			$html .= &add_form_form($words, $pos, $gloss, "really");
			$html .= $lookuphtml;
		} else {
			# everything fine - we can add the new synset
			$html .= &do_save_synset($pos, $words, $gloss);
		}	
	}
	$html .= qq|\t</div>|;
	$html .= qq|</div>|;
	return $html;
}

sub do_save_synset {
	my $pos = shift;
	my $words = shift;
	my $gloss = shift;

	my $html;

	# my @words = split(/[^\w\d_]/, $words);
	my @words = split(/\s*,\s*/, $words);
	my $word = shift(@words);
	$word =~ s/\s+/_/g;
	my $synset = $wn->new_synset("$word", $pos);
	$synset->gloss($gloss);
	foreach (@words) {
		s/\s+/_/g;
		$synset->add_words($_);
	}

	$wn->unlock;
	my $offset = $synset->write;
	$wn->lock;

	$html .= qq|<div class="info">
	The new Synset '$words' has been saved.<br/>
	Please <a href="$url?show=$offset">proceed</a> with adding Hypernyms, Hyponyms, ...
</div>|;
	return $html;
}

sub pointer_form {
	my $datatype = shift;
	my $offset = shift;
	my $query = shift || "";

	#my $wn = new Lingua::Wordnet("$conf{wndir}");

	my $synset = get_synset_by_offset($wn, $offset);

	if (!$synset) {
		return qq|<div class="error">No Synset found with offset '$offset'</div>|;
	} 

	my ($txt, $gloss) = get_text_gloss($synset);

	my $html;
	$html .= qq|<div class="box">\n|;
	$html .= qq|\t<div class="boxheader">\n|;
	$html .= qq|\t\t<div class="boxheaderoffset">$offset</div>\n|;
	$html .= qq|\t\tEdit $datatypes{$datatype} of <span class="word">$txt</span>\n|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t<br/>\n|;

	my @tmpsynsets;
	eval "\@tmpsynsets = \$synset->$datatype";
	if ($@) { die "eval error: $@"; }
	my %alreadylinked = map { $_->offset => 1 } @tmpsynsets;

	my $i = 0;
	if (scalar @tmpsynsets) {
		$html .= qq|\t<div class="boxsubheader">Already linked Synsets:</div>\n|;
		foreach my $syn (@tmpsynsets) {
			my ($words, $gloss) = get_text_gloss($syn);
			my $xoffset = $syn->offset;
			$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;
					
			$html .= qq|<div style="float:right;"><a href="$url?delpointer=$datatype;offset=$offset;deloffset=$xoffset"><img src="$conf{webdir}/images/del.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Delete $datatypes{$datatype} Pointer" title="Delete $datatypes{$datatype} Pointer"></a></div>|;

			$html .= qq|<a href="$url?show=$xoffset" class="word">$words</a> <span class="offset">($xoffset)</span>\n|;
			$html .= qq|<div class="glossary">$gloss</div>\n|;
			$html .= qq|</div>|;

		}
	}

	$html .= qq|\t<br/>\n|;
	$html .= qq|\t<div class="boxsubheader">Add a new $datatypes{$datatype}-Pointer to Synsets:</div>\n|;
	$html .= qq|\t<div class="boxtext">\n|;
	$html .= qq|\t$datatypes{$datatype}: $datatypedefinitions{$datatype}<br/><br/>\n|;
	$html .= qq|\t<form action="$url">
		<input type="hidden" name="offset" value="$offset"/>
		<input type="hidden" name="editpointer" value="$datatype"/>
		Search for Synset to add: <input type="text" name="q" value="| . encode_entities($query) . qq|"> <input type="submit" name="action_search" value="Search" class="formbutton">
	</form>\n|;
	$html .= qq|\t</div>\n|;
	
	if ($query) {
	
		# init spellchecker
		&initspeller();

		my @wordsorig = split(/\s+/, $query);
		my ($found, $synsets) = &get_synsets($wn, \@wordsorig);

		# sort synsets in order of orig words
		my $window_size = $#wordsorig;
		my %order;
		while ($window_size >= 0) {
			for (my $i = 0; $i <= $#wordsorig-$window_size; $i++) {
				my $tmp = join(" ", @wordsorig[$i..$i+$window_size]);
				$order{$tmp} = $i;
			}
			$window_size--;
		}

		if (scalar keys %{$synsets}) {

			foreach (sort { $order{$a} <=> $order{$b} } keys %{$synsets}) {
				next unless $synsets->{$_};
				(my $name = $_) =~ s/\s/_/g;


				$html .= qq|<div class="boxsubheader">Synsets of <span class="word">$name</word></div>\n|;
				foreach my $type (@types) {
					my @sets = grep { $_->[0] eq $type } @{$synsets->{$_}};
					if (scalar(@sets)) {
						$html .= qq|<div class="postype">$types{$type}:</div>\n|;
						my $i = 0;
						foreach my $synset (@sets) {
							my ($words, $gloss) = get_text_gloss($synset->[1]);
							my $xoffset = $synset->[1]->offset;
							$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;

							if ($alreadylinked{$xoffset}) {
								$html .= qq|<div style="float:right;"><a href="$url?delpointer=$datatype;offset=$offset;deloffset=$xoffset"><img src="$conf{webdir}/images/del.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Delete $datatypes{$datatype} Pointer" title="Delete $datatypes{$datatype} Pointer"></a></div>|;
							} elsif ($synset->[1]->{offset} eq $offset) {

							} else {
								$html .= qq|<div style="float:right;"><a href="$url?savepointer=$datatype;offset=$offset;addoffset=$xoffset"><img src="$conf{webdir}/images/plus.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Add $datatypes{$datatype} Pointer" title="Add $datatypes{$datatype} Pointer"></a></div>|;
							}
							$html .= qq|<a href="$url?show=$xoffset" class="word">$words</a> <span class="offset">($xoffset)</span>\n|;
							$html .= qq|<div class="glossary">$gloss</div>\n|;

							$html .= qq|</div>|;
						} # end @sets
					}
				} # end @types
				$html .= qq|<br/>|;
			} # end %synsets
		}


		if (scalar(@wordsorig) != scalar(keys %{$found})) {
			$html .= qq|<br/>\n<div class="boxsubheader">Words without Synsets</div>\n|;
			my $i = 0;
			foreach (@wordsorig) {
				unless ($found->{$_}) {
					(my $name = $_) =~ s/\s/_/g;
					$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;
					$html .= qq|<span class="word">$_</span>|;
					$html .= qq|</div>|;
				}
			}
		}
	}


	$html .= qq|\t<br/><div style="text-align: center"><a href="$url?show=$offset;open=$datatype#$datatype" class="formbutton">Back</a></div><br/>\n|;
#	$html .= qq|</div>\n|;
	$html .= qq|</div>\n|;
	return $html;
}

sub save_pointer {
	my $datatype = shift;
	my $offset = shift;
	my $addoffset = shift;

	#my $wn = new Lingua::Wordnet("$conf{wndir}");

	my $synset = get_synset_by_offset($wn, $offset);
	if (!$synset) {
		return qq|<div class="error">No Synset found with offset '$offset'</div>|;
	} 

	my $addsynset = get_synset_by_offset($wn, $addoffset);
	if (!$addsynset) {
		return qq|<div class="error">No Synset found with offset '$addoffset'</div>|;
	} 

	my $command = "add_" . $datatype;

	$wn->unlock;
	$synset->$command($addsynset);
	$synset->write;

	if (my $revpointer = $reversepointer{$datatype}) {
		my $revcommand = "add_" . $revpointer;	
		$addsynset->$revcommand($synset);
		$addsynset->write;
	}
	$wn->lock;

	return &pointer_form($datatype, $offset);
}


sub del_pointer {
	my $datatype = shift;
	my $offset = shift;
	my $deloffset = shift;

	#my $wn = new Lingua::Wordnet("$conf{wndir}");

	my $synset = get_synset_by_offset($wn, $offset);
	if (!$synset) {
		return qq|<div class="error">No Synset found with offset '$offset'</div>|;
	} 

	my $delsynset = get_synset_by_offset($wn, $deloffset);
	if (!$delsynset) {
		return qq|<div class="error">No Synset found with offset '$deloffset'</div>|;
	} 

	&do_delpointer($datatype, $synset, $delsynset);
	return &pointer_form($datatype, $offset);
}

sub do_delpointer {
	my $datatype = shift;
	my $synset = shift;
	my $delsynset = shift;

	my $command = "delete_" . $datatype;

	$wn->unlock;
	$synset->$command($delsynset);
	$synset->write;

	if (my $revpointer = $reversepointer{$datatype}) {
		my $revcommand = "delete_" . $revpointer;	
		$delsynset->$revcommand($synset);
		$delsynset->write;
	}
	$wn->lock;
}

##### words and glossary
sub edit_form {
	my $offset = shift || "";
	my $words = shift || "";
	my $pos = shift || "n";
	my $gloss = shift || "";
	my $error = shift;
	
	my $synset;

	if (!$offset || $offset eq "new") {
		$synset = $wn->new_synset("", "");
	} else {
		$synset = get_synset_by_offset($wn, $offset);
		if (!$synset) {
			return qq|<div class="error">No Synset found with offset '$offset'</div>|;
		} 
	}

	my $txt;
	my @words;
	if ($words) {
		@words = @{$words};
		$txt = join(", ", @words);
	} else {
		($txt, $gloss) = get_text_gloss($synset);
		@words = $synset->words();
	}

	my $html;
	$html .= qq|<div class="box">\n|;
	$html .= qq|\t<div class="boxheader">\n|;
	$html .= qq|\t\t<div class="boxheaderoffset">$offset</div>\n|;
	if ($offset eq "new") {
		$html .= qq|Add a new Synset|;
	} else {
		$html .= qq|Edit Synset <span class="word">$txt</span>|;
	}
	$html .= qq|</div>\n|;
	$html .= qq|\t<form action="$url">
		<input type="hidden" name="edit" value="$offset"/>\n|;

	$html .= qq|\t\t<div class="error">$error</div>\n| if $error;

	$html .= qq|\t\t<div class="boxsubheader">Part-of-speech</div>\n|;
	$html .= qq|\t\t<div class="listrow0">|;
	$html .= qq|\t\t<select name="pos">\n|;
	foreach my $p (@types) {
		my $sel = ($pos eq $p) ? " selected " : "";
		$html .= qq|\t\t\t<option value="$p"$sel>$types{$p}</option>\n|;
	}
	$html .= qq|\t\t</select>\n|;
	$html .= qq|</div>\n|;

	$html .= qq|<br/>\n|;

	$html .= qq|\t\t<div class="boxsubheader">Words</div>\n|;

	my $i = 0;
	push @words, "" if ($#words == -1);
	foreach (@words) {
		s/_/ /g;

		$html .= qq|<div class="listrow| . ($i % 2) . qq|">|;
		$html .= qq|<input name="words[$i]" type="text" size="50" value="| . encode_entities($_) . qq|"/>|;

		# delete word
		if ($#words > 0) {
			$html .= qq|<input type="image" src="$conf{webdir}/images/del.png" border="0" name="action_delword_$i" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Delete" title="Delete" hspace="5">|;
		} else {
			$html .= qq|<img src="$conf{webdir}/images/trans.png"  border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="" hspace="5">|;
		}
		# move up and down
		if ($i > 0) {
			$html .= qq|<input type="image" src="$conf{webdir}/images/arrow_u.png" border="0" name="action_upword_$i" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Move Up" title="Move Up" hspace="5">|;
		} else {
			$html .= qq|<img src="$conf{webdir}/images/trans.png"  border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="" hspace="5">|;
		}
		if ($i < $#words) {
			$html .= qq|<input type="image" src="$conf{webdir}/images/arrow_d.png" border="0" name="action_downword_$i" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Move Up" title="Move Up" hspace="5">|;
		} else {
			$html .= qq|<img src="$conf{webdir}/images/trans.png"  border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="" hspace="5">|;
		}
		$html .= qq|</div>|;
		$i++;
	}

	$html .= qq|\t<div class="listrow| . ($i%2) . qq|">\n|;
	$html .= qq|\t\t Add a new Word <input type="image" src="$conf{webdir}/images/plus.png" border="0" name="action_addword" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Add Word" title="Add word" hspace="5" align="absmiddle"><br/>\n|;
	$html .= qq|\t</div>|;

	$html .= qq|\t<br/>\n|;
	$html .= qq|\t\t<div class="boxsubheader">Glossary</div>\n|;
	$html .= qq|\t\t<div class="listrow0">|;
	$html .= qq|<input type="text" class="inputbig" name="gloss" value="| . encode_entities($gloss) . qq|"/>|;
	$html .= qq|</div>\n|;

	$html .= qq|\t<div class="boxtext">\n|;
	$html .= qq|\t\t<div style="text-align:center">
			<a href="$url?show=$offset" class="formbutton">Back</a> <input type="submit" name="save" value="Save" class="formbutton">
		</div>
</form>|;
	$html .= qq|\t</div>\n|;
	$html .= qq|</div>\n|;

	return $html;
}

sub save {
	my $offset = shift;
	my $words = shift;
	my $pos = shift || "n";
	my $gloss = shift;

	# TODO add a new synset if there is no offset ...
	my $html; my $error;

	my $synset;

	if (!$offset || $offset eq "new") {
		$synset = $wn->new_synset("", "$pos");
	} else {
		$synset = get_synset_by_offset($wn, $offset);
	}


	my @tmp;
	# clean up words
	if ($words && ref($words) eq "ARRAY") {
		foreach (@{$words}) {
			s/^\s+//;
			s/\s+$//;
			s/\s/_/g;
			push @tmp, $_ if $_;
		}
	}
	# check if there is still at least one word left
	if ($#tmp == -1) {
		$error .= "At least one word must be defined!<br/>";	
	}

	# no error => save words and glossary
	unless ($error) {
		# set words
		$synset->words(@tmp);

		# set glossary
		my $newgloss = $synset->gloss($gloss);

		# unlock wn
		$wn->unlock;
		$offset = $synset->write;
		$wn->lock;

		my ($txt, $oldgloss) = get_text_gloss($synset);

		$html .= qq|<div class="box">\n|;

		$html .= qq|\t<div class="boxheader">\n|;
		$html .= qq|\t\t<div style="float:right; margin-left: 10px;"><a href="$url?edit=$offset"><img src="$conf{webdir}/images/edit.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Edit" title="Edit" align="absmiddle"></a></div>|;
		$html .= qq|\t\t<div class="boxheaderoffset">$offset</div>\n|;
		$html .= qq|\t\t\tSynset <span class="word">$txt</span> saved\n|;
		$html .= qq|\t</div>\n|;
		$html .= qq|\t<br/>|;

		$html .= qq|\t<div class="boxsubheader">\n|;
		$html .= qq|\t\tThe following words have been assigned to this synset:|;
		$html .= qq|\t</div>\n|;
		$html .= qq|\t<div class="boxtext">\n|;
		foreach ($synset->words) {
			$html .= qq|\t\t<div class="word">$_</div>\n|;
		}
		$html .= qq|\t<br/>\n|;
		
		$html .= qq|\t\tOld Words: <span class="old">$txt</span>|;
		$html .= qq|\t</div>\n|;

		$html .= qq|\t<br/>\n|;

		$html .= qq|\t<div class="boxsubheader">\n|;
		$html .= qq|\t\tThe following text has been assigned to the glossary of this synset:|;
		$html .= qq|\t</div>\n|;


		$html .= qq|\t<div class="boxtext">\n|;
		$html .= qq|\t\t<div class="word">$newgloss</div><br/>\n|;
		$html .= qq|\t\tOld Glossary: <span class="old">$oldgloss</span>|;
		$html .= qq|\t</div>\n|;
		$html .= qq|<br/>|;

		$html .= qq|\t<div class="boxsubheader">\n|;
		$html .= qq|\t\tPlease proceed now with adding pointers|;
		$html .= qq|\t</div>\n|;
		$html .= qq|<br/>|;
		$html .= qq|\t<div style="text-align: center">|;
		$html .= qq|<a href="$url?show=$offset" class="formbutton">Proceed</a>|;
		$html .= qq|</div>\n|;
		$html .= qq|\t<br/>|;

		$html .= qq|</div>\n|;
	} else {
		$html .= &edit_form($offset, \@tmp, $pos, $gloss, $error);
	}
	return $html;
}

sub del_form {
	my $offset = shift;

	my $synset = get_synset_by_offset($wn, $offset);

	if (!$synset) {
		return qq|<div class="error">No Synset found with offset '$offset'<br/>
	[<a href="javascript:history.back();">back</a>]
</div>|;
	}

	my ($txt, $gloss) = get_text_gloss($synset);

	my $html;
	$html .= qq|<div class="box">\n|;
	$html .= qq|\t<div class="boxheader">\n|;
	$html .= qq|\t\t<div class="boxheaderoffset">$offset</div>\n|;
	$html .= qq|\t\tDelete Synset <span class="word">$txt</span>\n|;
	$html .= qq|\t</div>\n|;

	$html .= qq|\t<form action="$url" method="POST">\n|;
	$html .= qq|\t<input type="hidden" name="dodel" value="$offset"/>\n|;

	$html .= qq|\t<div class="boxtext">\n|;
	$html .= qq|\t\t<div class="warning">\n|;
	$html .= qq|\t\t<br/>Do you really want to delete this synset? The synset and all Pointers to other synsets will be deleted!<br/>|;
	$html .= qq|\t\t</div>\n|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t<br/>\n|;
	$html .= qq|\t<div class="boxsubheader">\n|;
	$html .= qq|\tWord(s) of this synset|;
	$html .= qq|\t</div>\n|;
	my $i = 0;
	foreach ($synset->words) {
		$_ =~ s/_/ /g;
		$_ =~ s/\%\d+$//;
		$html .= qq|\t\t<div class="listrow| . ($i++%2) . qq|"><span class="word">$_</span></div>\n|;
	}

	$html .= qq|\t<br/>\n|;

	$html .= qq|\t<div class="boxsubheader">\n|;
	$html .= qq|\tGlossary of this synset|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t<div class="listrow0">$gloss</div>\n|;

	$html .= qq|\t<br/>\n|;


	$html .= qq|\t<div class="boxsubheader">\n|;
	$html .= qq|\tPointers of this synset|;
	$html .= qq|\t</div>\n|;

	foreach my $datatype (@datatypes) {
		my @synsets;
        eval "\@synsets = \$synset->$datatype";
        if ($@) { die "eval error: $@"; }
		my $syncount = scalar(@synsets);
		if ($syncount > 0) {
			my $i = 0;
			$html .= qq|<a name="$datatype"></a>\n|;
			$html .= qq|<div class="boxsubheader">\n|;
			$html .= qq|\t<div style="float:right">$syncount Synsets</div>\n|;
			$html .= qq|$datatypes{$datatype}|;
			$html .= qq| - <span class="small">$datatypedefinitions{$datatype}</span>| if ($datatypedefinitions{$datatype});
			$html .= qq|</div>\n|;
			foreach my $syn (@synsets) {
				my ($words, $gloss) = get_text_gloss($syn);
				my $offset = $syn->offset;
				$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;
				$html .= qq|<a href="$url?show=$offset" class="word">$words</a> <span class="offset">($offset)</span>\n|;
				$html .= qq|<div class="glossary">$gloss</div>\n|;
				$html .= qq|</div>|;
			}
			$html .= qq|<br/>|;
		}
	}

	$html .= qq|<br/>|;
	$html .= qq|\t\t<div style="text-align: center">\n|;
	$html .= qq|\t\t<a href="$url?show=$offset" class="formbutton">Back</a> <input type="submit" name="delete" value="Really Delete" class="formbutton">\n|;
	$html .= qq|\t\t</div>\n|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t</form>\n|;
	$html .= qq|</div>\n|;

	return $html;
}


sub delete {
	my $offset = shift;

	my $synset = get_synset_by_offset($wn, $offset);

	if (!$synset) {
		return qq|<div class="error">No Synset found with offset '$offset'<br/>
	[<a href="javascript:history.back();">back</a>]
</div>|;
	}

	my ($txt, $gloss) = get_text_gloss($synset);

	my $html;
	$html .= qq|<div class="box">\n|;
	$html .= qq|\t<div class="boxheader">\n|;
	$html .= qq|\t\t<div class="boxheaderoffset">$offset</div>\n|;
	$html .= qq|\t\tDeleted Synset <span class="word">$txt</span>\n|;
	$html .= qq|\t</div>\n|;

	$html .= qq|\t<div class="boxtext">\n|;
	$html .= qq|\t\t<div class="info">\n|;
	$html .= qq|\t\t<br/>The Synset <span class="word">$txt</span> ($offset) has been deleted!<br/>|;
	$html .= qq|\t\<div class="glossary">$gloss</div>\n|;
	$html .= qq|\t\t</div>\n|;
	$html .= qq|\t</div>\n|;
	$html .= qq|\t<br/>\n|;

	foreach my $datatype (@datatypes) {
		my @synsets;
        eval "\@synsets = \$synset->$datatype";
        if ($@) { die "eval error: $@"; }
		my $syncount = scalar(@synsets);
		if ($syncount > 0) {
			foreach my $syn (@synsets) {
				&do_delpointer($datatype, $synset, $syn);
			}
		}
	}

	$synset->delete;

	$html .= qq|</div>\n|;

	return $html;

}

sub search {
	my $q = shift;

	# init a WN instance
	#my $wn = new Lingua::Wordnet("$conf{wndir}");
	# init spellchecker
	&initspeller();

	my $html;


	my @wordsorig = split(/\s+/, $q);
	my ($found, $synsets) = &get_synsets($wn, \@wordsorig);

	if (scalar(@wordsorig) == scalar(keys %{$found})) {
		my @keys = keys %{$synsets};
		if ($#keys == 0) {
			my @syns = @{$synsets->{$keys[0]}};
			if ($#syns == 0) {
				# only one synset found -> display details
				$html .= &show_synset($syns[0]->[1]->offset);
			}
		}
	}

	$html .= qq|<div class="box">\n|;
	# sort synsets in order of orig words
    my $window_size = $#wordsorig;
    my %order;
    while ($window_size >= 0) {
        for (my $i = 0; $i <= $#wordsorig-$window_size; $i++) {
            my $tmp = join(" ", @wordsorig[$i..$i+$window_size]);
            $order{$tmp} = $i;
        }
        $window_size--;
    }

	if (scalar keys %{$synsets}) {

		foreach (sort { $order{$a} <=> $order{$b} } keys %{$synsets}) {
			next unless $synsets->{$_};
			(my $name = $_) =~ s/\s/_/g;

			$html .= qq|<div class="boxsubheader">Synsets of <span class="word">$name</word></div>\n|;
			foreach my $type (@types) {
				my @sets = grep { $_->[0] eq $type } @{$synsets->{$_}};
				if (scalar(@sets)) {
					$html .= qq|<div class="postype">$types{$type}:</div>\n|;
					my $i = 0;
					foreach my $synset (@sets) {
						my ($words, $gloss) = get_text_gloss($synset->[1]);
						my $offset = $synset->[1]->offset;
						$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;
					
						$html .= qq|<div class="editbutton"><a href="$url?show=$offset"><img src="$conf{webdir}/images/arrow_r.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Edit" title="Edit"></a></div>|;

						$html .= qq|<a href="$url?show=$offset" class="word">$words</a> <span class="offset">($offset)</span>\n|;
						$html .= qq|<div class="glossary">$gloss</div>\n|;

						$html .= qq|</div>|;
					} # end @sets
				}
			} # end @types
			$html .= qq|<br/>|;
		} # end %synsets
	}


    if (scalar(@wordsorig) != scalar(keys %{$found})) {
        $html .= qq|<br/>\n<div class="boxsubheader">No Synsets found </div>\n|;
		my $i = 0;
        foreach (@wordsorig) {
            unless ($found->{$_}) {
                (my $name = $_) =~ s/\s/_/g;
				$html .= qq|<div class="listrow| . ($i++%2) . qq|">\n|;
				$html .= qq|<div style="float:right"><a href="$url?edit=new;words[0]=| . encode_entities($_) . qq|">New Synset with word '<span class="word">$_</span>'<img src="$conf{webdir}/images/plus.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Add a Synset" title="Add a Synset" align="absmiddle" hspace="5"/></a></div>|;
                $html .= qq|<span class="word">$_</span>|;
				$html .= qq|</div>|;
            }
        }
    }

	$html .= "</div>\n";
	return $html;
}

sub searchform {
	my $q = shift || "";

	my $html = qq|
<div class="searchform">
	<div style="float:right;">
		<a href="$url?edit=new">Add a new Synset <img src="$conf{webdir}/images/plus.png" border="0" width="$conf{iconwidth}" height="$conf{iconheight}" alt="Add a Synset" title="Add a Synset" align="absmiddle" hspace="5"/></a>
	</div>
	<form method="POST" action="$url" name="searchform">
		<input type="text" name="q" value="| . encode_entities($q) . qq|"> <input type="submit" name="action_search" value="Search" class="formbutton">
	</form>
</div>
|;
	
}

sub head {
	my $html = qq|
<html>
<head>
	<title>Lingua::Wordnet Editor</title>|;
	$html .= qq|<link rel="stylesheet" type="text/css" href="$conf{webdir}/wneditor.css"/>|;
	$html .= qq|</head>
<body>
<div class="masterdiv">
	<div class="mastertitle">Lingua::Wordnet Editor</div>
|;
	return $html;
}

sub foot {
	return qq|
	<div class="footer">
		(c) 2003 by <a href="mailto:koppi\@iua.upf.es">Koppi</a> (<a href="http://www.iua.upf.es/mtg/eng/" target="extern">MTG</a> / <a href="http://www.upf.es/" target="extern">Universitat Pompeu Fabra</a>)<br/>
		Lingua::Wordnet (c) <a href="http://www.brians.org/" target="extern">Dan Brian</a>
	</div>
</div>
</html>
|;
}
################################################################################
# WordNet related stuff
################################################################################

sub get_text_gloss {
	my $synset = shift;
	return ("", "") unless $synset;

	my @words = $synset->words;
	foreach (@words) {
		s/_/ /g;
		s/\%\d+//;
	}
	my $txt = join(", ", @words);

	my $gloss = $synset->gloss;

	return ($txt, $gloss);
}

sub get_synsets {
	my $wn  = shift;
    my $words = shift;
   
    my @words = @{$words};
	
    my %syns = ();
    my %found = ();
    my $window_size = $#words;
    
    while ($window_size >= 0) {
        for (my $i = 0; $i <= $#words-$window_size; $i++) {
            my $tmp = join(" ", @words[$i..$i+$window_size]);
            if (my $syn  = &has_synset($wn, $tmp)) {
                $syns{$tmp} = $syn;
                $found{$_}++ for split(/ /, $tmp);
                my @tmp = splice(@words, $i, $window_size+1);
				if ($#words != -1) {
					my ($tmpfound, $tmpsets) = &get_synsets($wn, \@words);
					foreach (keys %{$tmpsets}) {
						$syns{$_} = $tmpsets->{$_};
					}
					foreach (keys %{$tmpfound}) {
						$found{$_}++;
					}
				}
            }
        }
        $window_size--;
    }
    return (\%found, \%syns);
}

sub has_synset {
	my $wn = shift;
    my $text = shift;
    my $syn = &synset($wn, $text);
#	print "$text => " . $syn->[0]->[1] . "<br>";
    if (!$syn->[0]) {
        my @tmp = split(/\s+/, $text);
        foreach (@tmp) {
            if (my $x = (stemandspell($_))->[0]) {
                $_ = $x;
            }
        }
        my $t = join(" ", @tmp);
        $syn = &synset($wn, $t);
    }
    return ($syn->[0]->[1]) ? $syn : 0;
}


sub synset {
	my $wn = shift;
	my $text = shift;
	my $pos = shift;

	$text = lc($text);
	$text =~ s/\s+/_/g;
	my @synset;
	my @types = keys %types; # ('n', 'v', 'a', 's', 'r');
	# my @types = ('n');
	my @text = stemandspell($text);
	my %offsets;
	foreach my $t (@types) {
		foreach my $txt (@text) {
			foreach ($wn->lookup_synset($txt, $t)) {
				if (!$offsets{$_->offset}) {
					push @synset, [$t, $_];
					$offsets{$_->offset} = 1;
				}
			}
		}
	}
	return \@synset;
}

sub get_synset_by_offset {
	my $wn = shift;
	my $offset = shift;

	my $syn;
	eval {$syn = $wn->lookup_synset_offset($offset);};

	return $syn;
}

################################################################################
# Lingua related stuff (spelling, ...)
################################################################################
sub initspeller {

	$stemmer = new Lingua::Stem(-locale => 'EN-UK');
	$stemmer->stem_caching({ -level => 2 });

}

sub stem {
	my @words = @_;

	my $stemmed = $stemmer->stem(@words);

	return wantarray ? @{$stemmed} : $stemmed;
}

sub stemandspell {
	my @words = @_;

	my $debug = 1;

	my @spelled = spellchecker(@words);
	my @stemmed = stem(@words);
	my @morphed;
	foreach (@words) {
		push @morphed, morph($_);
	}
	my @morphspelled = spellchecker(@morphed);
	my @stemmedspelled = spellchecker(@stemmed);

	my %suffixes = (
		"s"	=> "",
		"ies"	=> "y",
		"es"	=> "e",
		"ed"	=> "e",
		"ing"	=> "e",
	);

	my @return;
	my %return;
	foreach my $i (0..$#words) {
		$return{$words[$i]}++;
	}
	foreach my $i (0..$#morphed) {
		if ($morphed[$i] eq $morphspelled[$i]) {
			$return{$morphed[$i]}++;
		}
	}
	@return = keys %return;
	return wantarray ? @return : \@return;
}

sub spellchecker {
	my @words = @_;

	my %incorr = (
		miss	=> 1,
		guess	=> 1,
		none	=> 1,
	);

	my @checked;
	foreach (@words) {
		my $err;
		my $guess;
		for my $r ( spellcheck($_) ) {
#			if ( $r->{'type'} eq 'ok' ) {
#				print "'$r->{'term'}' was found in the dictionary.\n";
#			}
#			print STDERR "**** " . $r->{'type'} . "\n";
			if ($incorr{$r->{'type'}}) {
#				print STDERR "*** >>$r->{'term'}<< was not found in the dictionary ($r->{'type'})\n";
				$err = "*** >>$r->{'term'}<< was not found in the dictionary\n";
			}
			if ($r->{'type'} eq "miss") {
#				print STDERR "    Near misses: " . join(", ", @{$r->{'misses'}}) . "\n";
				$err .= "Near misses: " . join(", ", @{$r->{'misses'}}) . "\n";
#				$guess = $r->{'misses'}->[0];
			} elsif ( $r->{'type'} eq "guess") {
#				print STDERR "*****    Root/affix guesses: " . join(", ", @{$r->{'guesses'}}) . "\n";
				$err .= "Root/affix guesses: " . join(", ", @{$r->{'guesses'}}) . "\n";
#				$guess = $r->{'guesses'}->[0];
			} 
#			elsif ( $r->{'type'} eq "none" ) {
#				$spellerr{$r->{term}} = $r->{'type'};
#				my $term = addtodict($r->{term});
#				$keywords{$term} += delete $keywords{$_};
#				delete $keywords{$_};
#			}
		}
		if ($err) { 
			push @checked, $err;
		} else {
			push @checked, $_;
		}
#		print "\n";
	}
	return wantarray ? @checked : \@checked;
}


## original code from WordNet-2.0/src/lib/morph.c
sub morph {
	my $word = shift;
	my @morphed;
	my @sufx = (
		# Noun suffixes
		"ses", "xes", "zes", "ches", "shes", "men", "ies", "s",
		# Verb suffixes 
		"ies", "es", "es", "ed", "ed", "ing", "ing", "s",
		# Adjective suffixes 
		"er", "est", "er", "est"
	);

	my @addr = (
		# Noun endings 
		"s", "x", "z", "ch", "sh", "man", "y", "",
		# Verb endings
		"y", "e", "", "e", "", "e", "", "",
		# Adjective endings
		"", "", "e", "e"
	);	

	foreach my $i(0..$#sufx) {
		if ($word =~ /$sufx[$i]$/) {
			(my $w = $word) =~ s/$sufx[$i]$/$addr[$i]/;
			push @morphed, $w;
#			return $word;
		}
	}
	return @morphed if scalar(@morphed);
	return ($word);
}



################################################################################
# Helper Functions (CGI, debug, ...)
################################################################################

#######################################################################
# getParams( CGI->query )                                             #
# read all params of CGI->$query and set variables named $param       #
# with $params value in the scope of the calling script               #
#######################################################################

sub getParams {
	my %args = (
		QUERY => '',
		FORMAT => '',
		@_
	);

	my %in = ();				# hash containing name => value pairs
	unless ($args{QUERY}) {
		return (wantarray) ? %in : \%in;
	}

	my $query = $args{QUERY};
	my @params = $query->param;	# array containing names of all params

	foreach my $name (@params) {
		my $value = $query->param($name);

		$name =~ s/\.[xy]$//i;

		### param is an action
		if ($name =~ /^action_/) {
			$name =~ s/\.[xy]$//i;
			@{$in{action}} = split(/_/, $name);
			shift @{$in{action}};	
		### params is part of an array

		} elsif ($name =~ /^([\w\%]+)((?:(?:\[\d+\])|(?:\{[\w\%]+\}))+)$/) {	

			##### $1 ... "myarr"
			##### $2 ... "[1]{3}[4]"
			dprint ("$1 ......... $2") if DEBUG;
			my $arr = $1;
			my $tmp = $2;
			#### hash keys that start with a \d have to be quoted,
			#### because perl interpretes the string 1_123 as
			#### the numerical value 1123
			$tmp =~ s/\{(\d[^}]*)\}/\{\"$1\"\}/g;
			my $foo = "\$in{$arr}" . "$tmp" ;
			if (!ref $value) {
				$value = chompData($value);
				$foo .=  " = \"" . quotemeta($value) . "\""; ## \"$value\"";
				dprint ("--- $foo") if DEBUG;
			} else {
				#### now image upload files are handled 'somehow'
				use Data::Dumper;
 				my $bar = Data::Dumper::Dumper($value);
				$bar =~ s/^\$VAR1 = //;
				$foo .= " = $bar";
				# print $foo;
			}
			### hit me baby one more time ###
			no strict "refs";
			### *ouch* ###
			eval  "$foo" ; 
			### satan ###
		### param is just a scalar
		} else {							
			$value = chompData($value);
			$in{$name} = $value;
		}
	}

	#### eine kleine idee zur erleichterung der actions 
	if (defined $in{action} and (ref($in{action}) eq 'ARRAY')) {
		$in{ACTION_NAME} = $in{action}->[0] || "";
		@{$in{ACTION_ARGV}} = @{$in{action}};
		shift(@{$in{ACTION_ARGV}});
		$in{ACTION_ARGC} = ($#{$in{ACTION_ARGV}} + 1) || 0;
		if (DEBUG) {
			dprint ("[ACTION_NAME] $in{ACTION_NAME}<br>") if DEBUG;
			dprint ("[ACTION_ARGV] (" . join(", ", @{$in{ACTION_ARGV}}) . ")") if DEBUG;
			dprint ("[ACTION_ARGC] $in{ACTION_ARGC} Werte") if DEBUG;
		}
	}

	return  (wantarray) ? %in : \%in;
}


### from Utils.pm
sub chompData {
    my $data = shift;
    return unless (defined $data );

    if (! ref $data) {
        chomp $data;
        return $data;
    }

    elsif (ref $data eq "ARRAY") {
        foreach my $i (0..$#{$data}) {
            ${$data}[$i] = &chompdata(${$data}[$i]);
        }
        return $data;
    }

    elsif (ref $data eq "HASH") {
        foreach my $k (keys %{$data}) {
            ${$data}{$k} = &chompdata(${$data}{$k})
        }
        return $data;
    } else {
        ### Filehandles und so schmaehs ...
		return $data;
	}
}

sub dprint
{
	my $d = shift;

 	my ($package, $filename, $line) = caller;
	my $info;
	my $dest = "STDERR";
	($info = $filename) =~ s/^.*\/([^\/].*)$/$1/;
	$info .= " || $line ";
	$dest ||= "STDOUT";
	my $oldfh = select($dest); $|++;
	print "[ $info ] ";
	if (ref $d) {
		print Dumper ($d);
	} else {
		print $d;
	}
	print "\n";
	select ($oldfh);
}

1;



=head1 DESCRIPTION


=head1 FUNCTIONS

 
=head1 HELPER FUNCTIONS


=over 4


=item B<getParams>

Reads Params from CGI query and returns %in or \%in if called in scalar context

Allows funny data structures, which can be used like this: 
<input type="text" name="array[0][1]" value="string">
=> $in{array}[0][1] = 'string'

If U wanna use these funny structures your param-names have 
to match

C</^(\w+)((?:(?:\[\d+\])|(?:\{\w+\}))+)$/>


Returns a hash with all your cgi parameters. Additionally, the following
keys are defined (these keys are to be consideres as reserved words, you
can't use them for your own values):
 C<ACTION_NAME> name of the action that should 
               be executed (scalar)
 C<ACTION_ARGV> arrayref to the arguments for 
               the action
 C<ACTION_ARGC> number of the arguments 
               in ACTION_ARGV

B<Required params:>
 QUERY  => $CGI::Object

=back

=cut



=head1 BUGS

None known yet ;-)

=head1 SEE ALSO

Used Perl Modules: L<CGI>, L<Lingua::Wordnet>, L<Lingua::Stem>, L<Lingua::Ispell>

WordNet: http://www.cogsci.princeton.edu/~wn/


=head1 AUTHORS

Markus 'Koppi' Koppenberger <koppi@iua.upf.es>

getParams(), dprint() and chompText() originally by slayven@xover.mud.at, 
sc@gimp.org, koppi@weirdoz.org

=cut

