#!/usr/bin/perl -w

use strict;

my $debug = 0;

my %defaults = (
	cgipath => "/var/www/cgi-bin/",
	webpath => "/var/www/html/",
	wndir => "/usr/local/WordNet-2.0/dict/lingua-wordnet/",
	ispell => "/usr/bin/ispell",
);

my %values;

print "
WNEDITOR - INSTALLER

This will install wneditor.pl to your cgi-bin directory and all
needed images and CSS to your webspace.

wneditor.pl is (c) 2003 by Markus 'Koppi' Koppenberger
Music Technology Group / Universitat Pompeu Fabra, Barcelona
Please send questions and comments to koppi\@iua.upf.es

";
sleep(1);

print "Checking availibility of needed Perl modules ...\n";

my @modules = qw(
	Lingua::Wordnet
	Lingua::Ispell
	Lingua::Stem
	CGI
	HTML::Entities
);

my @notok;
foreach (@modules) {
	printf ("  Checking %-20s", $_);
	eval " use $_; 1;";
	if ($@) {
		print "NOT OK\n";
		push @notok, $_;
	} else {
		print "OK\n";
	}
}
if (@notok) {
	print "\nThe following module(s) need to be installed to use wneditor\n";
	print join(", ", @notok);
	print "\n\n";
	print "You can get the from your nearest CPAN Mirror (http://cpan.org/SITES.html).\n";
	print "Or you may use 'perl -MCPAN -e shell' or 'cpan'.\n\n";
	exit;
} else {
	print "Modules seems to be alright, continuing installation ...\n\n";
}

print "Where do you want to install the Lingua::Wordnet editor?\n";
while (!$values{cgipath}) {
	print "Enter the full-path to your cgi-bin directory (ScriptAlias)
[$defaults{cgipath}]: ";
	$values{cgipath} = <STDIN>;
	chomp $values{cgipath};
	$values{cgipath} ||= $defaults{cgipath};
	if ($values{cgipath}) {
		$values{cgipath} = &checkpath("f", $values{cgipath}, "wneditor.pl");
	}
}

while (!$values{webpath}) {
	print "Enter the full-path to the directory where the images and css should be\n";
	print "installed, 'wneditor' will be appended (must be under your DocumentRoot!)\n";
	print "[$defaults{webpath}]: ";
	$values{webpath} = <STDIN>;
	chomp $values{webpath};
	$values{webpath} ||= $defaults{webpath};
	if ($values{webpath}) {
		$values{webpath} = &checkpath("d", $values{webpath}, "wneditor");
	}
}

print "\nNow we need some information for the configuration:\n\n";
while (!$values{wndir}) {
	print "Enter the full-path to the Lingua::Wordnet datadir\n";
	print "[/usr/local/WordNet-2.0/dict/lingua_wordnet/]: ";
	$values{wndir} = <STDIN>;
	chomp $values{wndir};
	$values{wndir} ||= "/usr/local/WordNet-2.0/dict/lingua-wordnet/";
	if ($values{wndir}) {
		$values{wndir} = &checkpath("d", $values{wndir}, "", "r");
		if (!-e "$values{wndir}/lingua_wordnet.data") {
			print "There is no Lingua::Wordnet data in '$values{wndir}'\n";
			$values{wndir} = "";
		}
	}
	
}

print "\nYour Lingua::Wordnet datafiles MUST be writable by the user under
which your cgi-scripts are executed! (Normally 'apache' or 'www' or 'www-data').\n\n";


($defaults{webdir} = $values{webpath}) =~ s/.*html//;
$defaults{webdir} .= "/" unless $defaults{webdir} =~ /\/$/;
$defaults{webdir} .= "wneditor";

while (!$values{webdir}) {
	print "Enter the url-path to the images and css\n";
	print "[$defaults{webdir}]: ";
	$values{webdir} = <STDIN>;
	chomp $values{webdir};
	$values{webdir} ||= $defaults{webdir};
}

while (!$values{ispell}) {
	print "Enter the full-path to your ispell binary\n";
	print "[/usr/bin/ispell]: ";
	$values{ispell} = <STDIN>;
	chomp $values{ispell};
	$values{ispell} ||= $defaults{ispell};
	if ($values{ispell}) {
		my @tmp = split(/\//, $values{ispell});
		my $ispellbin = pop @tmp;
		my $ispelldir = join("/", @tmp);
		$ispelldir = &checkpath("f", $ispelldir, $ispellbin, "r");
		if ($ispelldir) {
			$values{ispell} = $ispelldir . "/" . $ispellbin;
		} else {
			$values{ispell} = "";
		}
	}
}



print qq|
wneditor.pl will be created as '$values{cgipath}| . qq|wneditor.pl'
Images and CSS will be copied to '$values{webpath}| . qq|wneditor/'
Configuration:
  wndir => "$values{wndir}"
  webdir => "$values{webdir}"
  ispell => "$values{ispell}"

Do you want to continue? [Y/n] |;
my $ans = <STDIN>;
chomp $ans;
$ans ||= "y";
if ($ans =~ /y/i) {
	print "Installing wneditor ...\n";
	open (IN, "./cgi-bin/wneditor.pl") || die "Can't read ./cgi-bin/wneditor.pl $!";
	open (OUT, ">$values{cgipath}/wneditor.pl") || die "Can't write $values{cgipath}/wneditor.pl $!";
	while (<IN>) {
		if ($_ =~ /(\s+"wndir"\s+=>)\s+/) {
			print OUT "$1 '$values{wndir}',\n";
		} elsif ($_ =~ /(\s+"ispell"\s+=>)\s+/) {
			print OUT "$1 '$values{ispell}',\n";
		} elsif ($_ =~ /(\s+"webdir"\s+=>)\s+/) {
			print OUT "$1 '$values{webdir}',\n";
		} else {
			print OUT $_;
		}
	}
	close (IN);
	close (OUT);

	system ("cp -r ./html/* $values{webpath}/wneditor/") == 0 
		or die "Couldn't copy to $values{webpath}/wneditor/";

	(my $cgidir = $values{cgipath}) =~ s/.*cgi-bin//;
	$cgidir =~ s/\/\//\//g;
	print "Installation finished. Now you can point your webbrowser to the wneditor\n"; 
	print "(e.g. http://localhost/cgi-bin$cgidir" . "wneditor.pl)\n\n";

} else {
	print "Installation stopped ...\n";
}

sub checkpath {
	my $type = shift;
	my $path = shift;
	my $file = shift;
	my $readonly = shift;

	$path .= "/" unless $path =~ /\/$/;
	my $testfile = $path;
	$testfile .= ($file) ? $file : "x";
	if ($type eq "d" && $file && !-e $path.$file) {
		my $create = &createdir($path . $file);
		if ($create) {
			return $path;
		} else {
			return "";
		}
	} elsif ($type eq "d" && !$file) {


	} else {
		if (!-e $path) {
			my $create = &createdir($path);
			if ($create) {
				return $path;
			} else {
				return "";
			}
		} 
	}
	if (!-d $path) {
		print "$path is no directory!\n";
		return "";
	}
	if (!$readonly && !-w $testfile && !-w $path) {
		print "$testfile is not writable!\n";
		return "";
	}
	if ($file && !$readonly && -e "$path$file") {
		print "File '$path$file' already exists!\nDo you want to overwrite it? [N/y]: ";
		my $ans = <STDIN>;
		if ($ans =~ /y/i) {
			print "   OK, will overwrite '$path$file'. But don't blame me later!\n";
			print "   Press CTRL-C now to stop me ...\n";
			sleep(5);
		} else {
			return "";
		}
	}
	return $path;
}

sub createdir {
	my $dir = shift;

	print "Directory '$dir' does not exist.\nShall I create it? [Y/n] ";
	my $ans = <STDIN>;
	if ($ans =~ /y/i) {
		my @path = split(/\/+/, $dir);
		my $tmpdir = shift (@path);
		my $error;
		foreach (@path) {
			$tmpdir .= "/" . $_;
			if (-d $tmpdir) {
				next;
			} else {
				my $test = mkdir("$tmpdir");
				if (!$test) {
					print "    NOT OK, could not create $dir!\n";
					return 0;
				}
			}
		}
		print "    OK, directory $dir created!\n";
		return 1;
	} else {
		return 0;
	}
}

=comment - this would be nice, but i don't want to do it now ;-)

print "<debug> Searching for httpd.conf ...\n";
my @poshttpdconfs = qw(
	/etc/httpd/httpd.conf
	/etc/httpd/conf/httpd.conf
	/etc/apache/httpd.conf
);

my $httpdconf;
foreach (@poshttpdconfs) {
	if (-e $_) {
		$httpdconf = $_;
		last;
	}
}

my @scriptaliases;
if ($httpdconf) {
	print "<debug> Got httpd.conf: $httpdconf\n";
	if (-r $httpdconf) {
		@scriptaliases = get_scriptaliases($httpdconf);
	} else {
		print "<debug> $httpdconf found but not readable\n";
	}
} else {
	print "<debug> No httpd.conf found\n";
}

if ($#scriptaliases > 0) {
	print "Found more than one ScriptAlias. Choose one or enter a path:\n";
	my $i = 1;
	foreach (@scriptaliases) {
		print "[" . sprintf("% 2d", $i++) . "] $_->[1]\n";
	}
}


# try to guess script-aliases from httpd.conf
sub get_scriptaliases {
	my $httpdconf = shift;
	my @scriptaliases;
	open (F, "$httpdconf") || warn "Can't read $httpdconf $!";
	if (fileno(F)) {
		while (<F>) {
			chomp;
			if (/^\s*ScriptAlias\s+([^\s]+)\s+([^\s]+)/) {
				push @scriptaliases, [$1, $2];
			}
		}
		close (F);
	}
	return @scriptaliases;
}

=cut
